{% extends "base.html" %}
{% block title %}{{ i18n('edit_title', path=path) }}{% endblock %}
{% block links %}
<link rel="stylesheet" type="text/css" href="{{ static_url('edit.css') }}">
{%- endblock %}
{% block scripts %}
<script src="{{ static_url('common.js') }}"></script>
<script src="{{ static_url('edit.js') }}"></script>
{%- endblock %}
{% block content %}
<div id="pinned">
<noscript>
<div class="error">{{ i18n('edit_error_no_javascript') }}</div>
</noscript>
{%- if encoding %}
<div class="warning">{{ i18n('edit_warning_encoding', encoding=encoding) }}</div>
{%- endif %}
</div>
<div id="wrapper">
<textarea id="editor">{{ body }}</textarea>
<div id="toolbar">
  <div>
    <input id="btn-save" type="button" value="{{ i18n('button_save') }}" autocomplete="off" title="{{ i18n('button_save_tooltip', 'S') }}" accesskey="S">
    <input id="btn-exit" type="button" value="{{ i18n('button_exit') }}" autocomplete="off" title="{{ i18n('button_exit_tooltip', 'E') }}" accesskey="E">
  </div>
</div>
</div>
{% endblock %}